
#include <R.h>
#include <Rinternals.h>
#include <Rmath.h>

static double h(SEXP fun, SEXP state, SEXP otherArgs, SEXP rho);

static void proposal_setup(SEXP control, int d);

static void propose(SEXP state, SEXP proposal);

static SEXP getListElement(SEXP list, char *str);

static void spacing_setup(SEXP control);

static int space(void);

SEXP metropolis(SEXP fun, SEXP state, SEXP otherArgs, SEXP rho, SEXP nin,
    SEXP control)
{
     SEXP foo, result, resultnames, proposal, acceptance_rate, path;
     int i, k, n, d;
     double acceptances = 0.0;
     double tries = 0.0;

     double current_dens;

     if (! isFunction(fun))
         error("argument \"fun\" must be function");
     if (! isNumeric(state))
         error("argument \"state\" must be numeric");
     if (! isPairList(otherArgs))
         error("argument \"otherArgs\" must be pairlist");
     if (! isEnvironment(rho))
         error("argument \"rho\" must be environment");
     if (! isNumeric(nin))
         error("argument \"n\" must be numeric");
     if (! isVector(control))
         error("argument \"control\" must be list");
     if (LENGTH(nin) != 1)
         error("argument \"n\" must be scalar");
     PROTECT(foo = coerceVector(nin, REALSXP));
     n = REAL(foo)[0];
     UNPROTECT(1);

     d = LENGTH(state);

     proposal_setup(control, d);
     spacing_setup(control);

     PROTECT(result = allocVector(VECSXP, 2));
     PROTECT(acceptance_rate = allocVector(REALSXP, 1));
     SET_VECTOR_ELT(result, 0, acceptance_rate);
     PROTECT(path = allocMatrix(REALSXP, d, n));
     SET_VECTOR_ELT(result, 1, path);
     PROTECT(resultnames = allocVector(STRSXP, 2));
     SET_STRING_ELT(resultnames, 0, mkChar("accept"));
     SET_STRING_ELT(resultnames, 1, mkChar("path"));
     namesgets(result, resultnames);
     UNPROTECT(3);

     PROTECT(proposal = allocVector(REALSXP, d));

     GetRNGstate();

     current_dens = h(fun, state, otherArgs, rho);

     for (i = 0, k = 0; i < n; i++) {

         int j;
         double proposal_dens;
         int sp = space();

         for (j = 0; j < sp; j++) {

             if (current_dens <= 0.0)
                 error("nonpositive density at current state");

             propose(state, proposal);

             proposal_dens = h(fun, proposal, otherArgs, rho);

             if (proposal_dens < 0.0)
                 error("negative density at proposed state");

             if (proposal_dens > current_dens ||
                 unif_rand() < proposal_dens / current_dens) {

                 int jj;

                 for (jj = 0; jj < d; jj++)
                     REAL(state)[jj] = REAL(proposal)[jj];

                 current_dens = proposal_dens;

                 acceptances++;
             }

             tries++;
         }

         for (j = 0; j < d; j++)
             REAL(path)[k++] = REAL(state)[j];
     }

     PutRNGstate();

     REAL(acceptance_rate)[0] = acceptances / tries;

     UNPROTECT(2);
     return result;
}

static double h(SEXP fun, SEXP state, SEXP otherArgs, SEXP rho)
{
     SEXP call, result, foo;
     double bar;

     PROTECT(call = LCONS(fun, LCONS(state, otherArgs)));
     PROTECT(result = eval(call, rho));
     if (! isNumeric(result))
         error("result of function call must be numeric");
     if (LENGTH(result) != 1)
         error("result of function call must be scalar");
     PROTECT(foo = coerceVector(result, REALSXP));
     bar = REAL(foo)[0];
     UNPROTECT(3);
     return bar;
}

static double *scale_factor;
static double scale_factor_buffer;
static int scale_option = 0;
static int state_dimension;
#define CONSTANT   1
#define DIAGONAL   2
#define FULL       3

static void proposal_setup(SEXP control, int d)
{
    SEXP s = getListElement(control, "scaling");
    SEXP foo;

    state_dimension = d;

    if (s == R_NilValue)
        error("item \"scaling\" not found in \"control\" list.");
    if (! isNumeric(s))
        error("scaling must be numeric");
    PROTECT(foo = coerceVector(s, REALSXP));
    if (isMatrix(s)) {
        SEXP bar;
        PROTECT(bar = getAttrib(s, R_DimSymbol));
        if (INTEGER(bar)[0] == d && INTEGER(bar)[1] == d) {
            int i;
            scale_factor = (double *) R_alloc(d * d, sizeof(double));
            for (i = 0; i < d * d; i++)
                scale_factor[i] = REAL(foo)[i];
            scale_option = FULL;
        } else {
            error("dimensions of \"scaling\" matrix not d by d");
        }
        UNPROTECT(1);
    } else if (LENGTH(foo) == d) {
        int i;
        scale_factor = (double *) R_alloc(d, sizeof(double));
        for (i = 0; i < d; i++)
            scale_factor[i] = REAL(foo)[i];
        scale_option = DIAGONAL;
    } else if (LENGTH(foo) == 1) {
        scale_factor = &scale_factor_buffer;
        scale_factor[0] = REAL(foo)[0];
        scale_option = CONSTANT;
    } else {
        error("length of \"scaling\" vector not d or 1");
    }
    UNPROTECT(1);
}

static void propose(SEXP state, SEXP proposal)
{
    int d = state_dimension;
    int i, j, k;

    if (LENGTH(state) != d || LENGTH(proposal) != d)
        error("State or proposal length different from initialization\n");

    switch (scale_option) {
        case CONSTANT:
            for (j = 0; j < d; j++)
                REAL(proposal)[j] = REAL(state)[j]
                    + scale_factor[0] * norm_rand();
            break;
        case DIAGONAL:
            for (j = 0; j < d; j++)
                REAL(proposal)[j] = REAL(state)[j]
                    + scale_factor[j] * norm_rand();
            break;
        case FULL:
            for (j = 0; j < d; j++)
                REAL(proposal)[j] = REAL(state)[j];

            for (i = 0, k = 0; i < d; i++) {
                double u = norm_rand();
                for (j = 0; j < d; j++)
                    REAL(proposal)[j] += scale_factor[k++] * u;
            }
            break;
        default:
            error("bogus scaling option\n");
    }
}

/* Stolen from Section 4.6.6 of "Writing R Extensions" Version 1.2.1 */

static SEXP getListElement(SEXP list, char *str)
{
    SEXP elmt = R_NilValue;
    SEXP names = getAttrib(list, R_NamesSymbol);
    int i;

    for (i = 0; i < length(list); i++)
        if(strcmp(CHAR(STRING_ELT(names, i)), str) == 0) {
            elmt = VECTOR_ELT(list, i);
            break;
        }
    return elmt;
}

static int spacing = 0;
static double spacing_geo_prob = 0.0;
static int spacing_option = 0;
/* option 1 is CONSTANT, defined above */
#define GEOMETRIC  2

static void spacing_setup(SEXP control)
{
    SEXP s = getListElement(control, "spacing");
    SEXP bar, baz, qux;

    if (s == R_NilValue) {
        spacing_option = CONSTANT;
        spacing = 1;
        return;
    }

    if (! isVector(s))
        error("spacing must be a list");
    if (LENGTH(s) != 2)
        error("spacing must be a list with two elements");
    bar = VECTOR_ELT(s, 0);
    baz = VECTOR_ELT(s, 1);
    if (! isString(bar))
        error("first element of spacing list must be type character");
    if (! isNumeric(baz))
        error("second element of spacing list must be numeric");
    PROTECT(qux = coerceVector(baz, REALSXP));
    if (strcmp(CHAR(STRING_ELT(bar, 0)), "constant") == 0) {
        spacing_option = CONSTANT;
        spacing = REAL(qux)[0];
        if (spacing <= 0)
            error("non-positive constant spacing");
    } else if (strcmp(CHAR(STRING_ELT(bar, 0)), "geometric") == 0) {
        spacing_option = GEOMETRIC;
        spacing_geo_prob = REAL(qux)[0];
        if (spacing_geo_prob <= 0.0 || 1.0 <= spacing_geo_prob)
            error("geometric spacing probability not strictly between 0 and 1");
    } else {
        error("spacing type not recognized");
    }
    UNPROTECT(1);
}

static int space(void)
{
    switch (spacing_option) {
        case CONSTANT:
            return spacing;
        case GEOMETRIC:
            return 1 + rgeom(spacing_geo_prob);
        default:
            error("bogus spacing option\n");
    }
    /* NOT REACHED */
    return 0; /* just so gcc -Wall doesn't complain */
}

